/*
** util.c - written in milano by vesely on 6mar2013
** collected mail parsing utilities
*/
/*
* zdkimfilter - Sign outgoing, verify incoming mail messages

Copyright (C) 2013-2020 Alessandro Vesely

This file is part of zdkimfilter

zdkimfilter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

zdkimfilter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License version 3
along with zdkimfilter.  If not, see <http://www.gnu.org/licenses/>.

Additional permission under GNU GPLv3 section 7:

If you modify zdkimfilter, or any covered part of it, by linking or combining
it with OpenSSL, OpenDKIM, Sendmail, or any software developed by The Trusted
Domain Project or Sendmail Inc., containing parts covered by the applicable
licence, the licensor of zdkimfilter grants you additional permission to convey
the resulting work.
*/

#include <config.h>
#if !ZDKIMFILTER_DEBUG
#define NDEBUG
#endif

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#ifndef NO_IDN2
#include <idn2.h>
#endif
#include <arpa/nameser.h>
#include "util.h"
#include <libopendkim/dkim-mailparse.h>
#include <assert.h>

char *hdrval(const char *a, const char *b)
// b must be without trailing ':'
// return pointer after column if headers match, NULL otherwise
{
	assert(a && b && strchr(b, ':') == NULL);
	
	int c, d;
	do c = *(unsigned char const*)a++, d = *(unsigned char const*)b++;
	while (c != 0 && d != 0 && (c == d || tolower(c) == tolower(d)));
	
	if (d != 0 || c == 0)
		return NULL;

	while (c != ':')
		if (!isspace(c) || (c = *(unsigned char const*)a++) == 0)
			return NULL;

	return (char*)a;
}

int is_subdomain(char const *domain, char const *org_domain)
{
	if (!domain || !org_domain)
		return 0;

	unsigned dlen = strlen(domain);
	unsigned olen = strlen(org_domain);

	return dlen > olen && strincmp(domain + dlen - olen, org_domain, olen) == 0;
}

unsigned int longest_substring(char const *a, char const *b)
{
	assert(a);
	assert(b);

	unsigned int alen = strlen(a);
	unsigned int blen = strlen(b);
	unsigned int prev[blen], cur[blen];
	memset(cur, 0, sizeof cur);

	unsigned int longest = 0;
	for (unsigned int i = 0; i < alen; ++i)
	{
		memcpy(prev, cur, sizeof prev);
		for (unsigned int j = 0; j < blen; ++j)
		{
			if (a[i] == b[j])
			{
				if (i == 0 || j == 0)
					cur[j] = 1;
				else
					cur[j] = prev[j-1] + 1;

				if (cur[j] > longest)
					longest = cur[j];

			}
			else
				cur[j] = 0;
		}
	}

	return longest;
}

int utf8length(unsigned int first)
{
	if (first < 0xbfU) return -1;
	if (first < 0xdfU) return 2;
	if (first < 0xefU) return 3;
	if (first < 0xf8U) return 4;
	return -1;
}

uint16_t my_get16(const unsigned char *src)
{
	uint16_t dst;

	NS_GET16(dst, src);
	return dst;
}


#if !TEST_UTIL  && !TEST_LONGEST_SUBSTRING
#ifndef NO_IDN2
// these require more linking
static int is_ldh(char const *s)
{
	int dot = 0;
	int ch;
	while ((ch = *(unsigned const char*)s++) != 0)
		if (isalnum(ch) && ch != '-')
			dot = 0;
		else if (ch == '.' && dot == 0)
			++dot;
		else
		{
#if CONSIDER_UTF8_STRINGS
			int len = utf8length(ch);
			if (len > 0)
				while (--len > 0 &&
					((ch = *(unsigned const char*)s++) & 0xc0) == 0x80)
						continue;
			if (len != 0)
#endif // CONSIDER_UTF8_STRINGS
				return 0;
		}
	return dot == 0;
}

static int check_domain(char *domain, char** out)
/*
* Check and normalize domain.  Return 0 if ok, 1 if bad, -1 fatal.
*/
{
	assert(domain);
	assert(out);

	char *out2 = NULL;
	*out = NULL;
	int rtc = idn2_to_ascii_8z(domain, &out2,
		IDN2_NFC_INPUT | IDN2_NONTRANSITIONAL);
	if (rtc == IDN2_MALLOC)
		return -1;

	if (rtc == IDN2_OK)
	{
		if (is_ldh(out2))
		{
			rtc = idn2_to_unicode_8z8z(out2, out, 0);
			rtc = rtc == IDN2_OK? 0: rtc == IDN2_MALLOC? -1: 1;
		}
		else
			rtc = 1;
		free(out2);
		return rtc;
	}

//	(*do_report)(LOG_WARNING, "cannot convert normalize %s: %s (%s)\n",
//		domain, idn2_strerror (rtc), idn2_strerror_name(rtc));
	return 1;
}

static inline int x2i(int ch)
{
	assert(isxdigit(ch));

	if (ch <= '9') return ch - '0';
	if (ch <= 'F') return ch - 'A' + 10;
	return ch - 'a' + 10;
}

static void urldecode_inplace(char *s)
{
	assert(s);
	char *o = s;
	int ch;
	while ((ch = *(unsigned char*)s++) != 0)
	{
		int ch1, ch2;

		if (ch == '%' &&
			isxdigit(ch1 = ((unsigned char*)s)[0]) &&
			isxdigit(ch2 = ((unsigned char*)s)[1]))
		{
			*o++ = 16*x2i(ch1) + x2i(ch2);
			s += 2;
		}
		else // silently pretend '%' is valid
			*o++ = ch;
	}
	*o = 0;
}

int mailto_domain(char const *s, char**out)
/*
* Find the first valid domain in a mailto: link of a List- header field.
* Return 0 if found, and out is the normalized domain.  Return 1 if no
* valid domain is found, -1 if a hard error occurred.
*
* mailto url: https://datatracker.ietf.org/doc/html/rfc6068#section-2
* List-* fields: https://datatracker.ietf.org/doc/html/rfc2369#section-3
*/
{
	assert(s);
	assert(out);

	char *scrap = strdup(s);
	if (scrap == NULL)
		return -1;

	char *sc = scrap;
	int rtc = 1;
	*out =NULL;
	while (rtc == 1)
	{
		// .......                 012345678
		char *mailto = strstr(sc, "<mailto:");
		if (mailto == NULL)
			break;

		mailto += 8;
		sc = mailto;
		char *e = (char*)
			my_mail_matching_paren(mailto, mailto + strlen(mailto), '<', '>');
		if (e && *e == '>')
		{
			*e = 0;
			sc = e + 1;

			char *query = strchr(mailto, '?');
			if (query)
				*query = 0;
			char *next = mailto;
			while (next)
			{
				char *comma = strchr(next, ',');
				if (comma)
					*comma++ = 0;

				char *domain = strchr(next, '@');
				next = comma;
				if (domain)
				{
					domain += 1;
					urldecode_inplace(domain);
					if ((rtc = check_domain(domain, out)) != 1)
						break;
				}
			}
		}
	}

	free(scrap);
	return rtc;
}
#endif // NO_IDN2

#if TEST_MAILTO
#include <stdio.h>

int main(int argc, char *argv[])
{
	if (argc == 2 && strcmp(argv[1], "--version") == 0)
		printf("%s version 1\n", argv[0]);
	else
		for (int i = 1; i < argc; ++i)
		{
			char *domain;
			int rtc = mailto_domain(argv[i], &domain);
			printf("%d %s\n", rtc, rtc == 0? domain: "/");
			if (rtc == 0)
				free(domain);
		}
	return 0;
}

#endif // TEST_MAILTO
#endif // !TEST_UTIL  && !TEST_LONGEST_SUBSTRING

char const *skip_token(char const *s)
{
	int ch;
	while ((ch = *(unsigned char*)s) > 32 &&
		ch < 128 &&
		strchr("()<>@,;:\\\"/[]?=", ch) == NULL)
			++s;
	return s;
}

char *skip_comment(char const *s)
{
	assert(s && *s == '(');
	
	int comment = 1;
	for (;;)
	{
		switch (*(unsigned char*)++s)
		{
			case 0:
				return NULL;

			case '(':
				++comment;
				break;

			case ')':
				if (--comment <= 0)
					return (char*)s;
				break;

			case '\\': // quoted pair, backslash cannot be last char
				++s;    // since there must be a newline anyway
				break;

			default:
				break;
		}
	}
}

char *skip_cfws(char const *s)
{
	while (s)
	{
		int ch;
		while (isspace(ch = *(unsigned char const*)s))
			++s;
		if (ch == '(')
		{
			if ((s = skip_comment(s)) != NULL)
			{
				assert(*s == ')');
				++s;
			}
		}
		else if (ch)
			break;
		else
			s = NULL;
	}
	return (char*)s;
}

// moved from myadsp.c 17 Feb 2021
char *skip_fws(char const *s)
{
	if (s)
	{
		int ch;
		while (isspace(ch = *(unsigned char*)s))
			++s;
		if (ch == 0)
			s = NULL;
	}
	return (char*) s;
}

char *skip_cfws_sep(char const *s, int sep)
{
	char *p = skip_cfws(s);
	if (*(unsigned char *)p != sep)
		return NULL;

	return skip_cfws(p+1);
}

static char *skip_digits(char const *s)
{
	while (s)
	{
		int ch;
		while (isdigit(ch = *(unsigned char const*)s))
			++s;
		if (ch == '(')
		{
			if ((s = skip_comment(s)) != NULL)
			{
				assert(*s == ')');
				++s;
			}
		}
		else if (ch)
			break;
		else
			s = NULL;
	}
	return (char*)s;
}

typedef struct token
{
	char *p, *q;
	int end_delimiter;
} token;

static char* a_r_scan(token *tok, int joint)
/*
* Tokenize the string as a sequence of 0-terminated words and return the type.
* On first call, p points to the input string to be parsed and q == NULL.
* The input string is overwritten with the output as the scan proceeds.
* On subsequent calls, p and q point to the input and output respectively.
* The start of the output is the return value.
*/
{
	assert(tok && tok->p && tok->q <= tok->p);

	char *p = skip_cfws(tok->p);
	char *q = tok->q? tok->q: tok->p, *entry = q;

	int quot = 0, escape = 0, ch;

	while (p && (ch = *(unsigned char*)p++) != 0)
	{
		assert (q <= p);

		if (escape)
		{
			*q++ = ch;
			escape = 0;
			continue;
		}

		if (quot)
		{
			switch (ch)
			{
				case '\\':
					escape = 1;
					continue;

				case '"':
					quot = 0;
					continue;

				default:
					*q++ = ch;
					continue;
			}
		}

		switch (ch)
		{
			case '\\':
				escape = 1;
				continue;

			case '"':
				quot = 1;
				continue;

			case '(':
				p = skip_comment(p - 1);
				if (p)
				{
					assert(*p == ')');
					++p;
				}
				continue;

			default:
				break;
		}

		if (isspace(ch))
		{
			char *next = skip_cfws(p);

			if (next && joint == 0) // authserv-id: discard version
			{
				next = skip_digits(next);
				if (next)
					next = skip_cfws(next);
			}

			int ch2 = next? *(unsigned char*)next: 0;

			/*
			* If a joint is admitted, CFWS can be placed around it
			*/
			if (joint && ch2 == joint && (next = skip_cfws(next + 1)) != 0)
			{
				*q++ = joint;
				p = next;
				continue;
			}

			/*
			* If the next value is a delimiter, use it.
			* Otherwise space or zero is the delimiter.
			*/
			if (ch2 && strchr(";=", ch2) != NULL)
			{
				ch = ch2;
				p = next + 1;
			}
			else if (ch2)
				ch = ' ';
			else
				ch = 0;
			break;
		}
		else if (strchr(";=", ch) != NULL)
			break;

		*q++ = ch;
	}

	if (p == NULL || escape || quot)
		return NULL;

	*q++ = 0;
	tok->q = q;
	tok->p = p;
	tok->end_delimiter = ch;

	assert(tok && tok->p && tok->q <= tok->p);
	return entry;
}

int
a_r_parse(char const *a_r, int (*cb)(void*, int, name_val*, size_t), void *cbv)
/*
* Parse a_r, which should start with the authserv-id, and call back cb with
* arguments:
*
* 1 (void*) the cbv given on entry,
* 2 (int) -1 for authserv-id, 0, 1, ... for "resinfo" stanzas, rtc for last call
* 3 (name_val*) an array of name=value pairs, and
* 4 (size_t) the number of elements in the array.
*
* The value is null for authserv-id.  The array itself is null on the last call.
*/
{
	assert(a_r);
	assert(cb);

	char *s = strdup(a_r);
	name_val resinfo[16];
	int rtc = 0;

	if (s == NULL) return -1;

	enum a_r_state {
		a_r_server,
		a_r_method,
		a_r_name,
		a_r_value
	} state = a_r_server;

	// Some of the elements that expect an '=' have an optional joint.
	static const int joint[] = {
		0,   // a_r_server, [CFWS] authserv-id [ CFWS authres-version ]
		'/', // a_r_method, Keyword [ [CFWS] "/" [CFWS] method-version ]
		'.', // a_r_name, ptype [CFWS] "." [CFWS] property [CFWS]
		'@'  //a_r_value, ( value / [ [ local-part ] "@" ] domain-name )
	};

	int count = 0;
	size_t n = 0;
	token tok;
	memset(&tok, 0, sizeof tok);
	tok.p = s;

	do
	{
		char *r = a_r_scan(&tok, joint[state]);
		if (r == NULL)
		{
			rtc = -1;
			break;
		}

		switch(state)
		{
			case a_r_server:
				if (tok.end_delimiter == ';')
				{
					resinfo[0].name = r;
					resinfo[0].value = NULL;
					rtc = (*cb)(cbv, -1, resinfo, 1);
					state = a_r_method;
				}
				break;

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wimplicit-fallthrough="
			case a_r_method:
				n = 0;
				if (count == 0 && stricmp(r, "none") == 0)
					break;

				/* else thru */

			case a_r_name:
				resinfo[n].name = r;
				state = a_r_value;
				if (tok.end_delimiter != '=')
					rtc = -2;
				break;
#pragma GCC diagnostic pop

			case a_r_value:
				resinfo[n].value = r;
				++n;

				if (tok.end_delimiter == ' ')
					state = a_r_name;

				else if (tok.end_delimiter == ';' || tok.end_delimiter == 0)
				{
					state = a_r_method;
					rtc = (*cb)(cbv, count, resinfo, n);
					++count;
				}
				else
					rtc = -3;
				break;

			default:
				assert(0);
				break;
		}
	} while (rtc == 0 && tok.end_delimiter != 0 &&
		n < sizeof resinfo / sizeof resinfo[0]);

	if (rtc == 0 && tok.end_delimiter != 0) rtc = -4;
	rtc = (*cb)(cbv, rtc, NULL, 0); // last call

	free(s);
	return rtc;
}

#if defined TEST_UTIL
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

static int verbose = 0;

static int my_cb(void *v, int step, name_val* nv, size_t nv_count)
{
	(void)v;
	assert(v == NULL);

	if (verbose)
		printf("%3zu value%s at step %d\n", nv_count, nv_count > 1? "s": "", step);

	if (nv == NULL)
		return step;

	if (step < 0)
		printf("%s;\n", nv[0].name);
	else
	{
		for (size_t i = 0; i < nv_count; ++i)
		{
			if (verbose)
				printf(" name/value [%zu]: ", i);
			printf(" %s=%s", nv[i].name, nv[i].value);
			if (verbose)
				putchar('\n');
		}
		putchar('\n');
	}

	return 0;
}

int main(int argc, char *argv[])
{
	char *fname = NULL;
	int i, errs = 0, rtc = 1;

	for (i = 1; i < argc; ++i)
	{
		char *arg = argv[i];

		if (arg[0] == '-' && arg[1])
		{
			int ch;
			while ((ch = *++arg) != 0)
			{
				switch (ch)
				{
					case 'v':
						++verbose;
						break;

					default:
						fprintf(stderr, "Invalid arg[%d]: %s\n", i, argv[i]);
						++errs;
						break;
				}
			}
		}
		else if (fname == NULL)
			fname = arg;
		else
		{
			fprintf(stderr, "Unexpected arg[%d]: %s\n", i, argv[i]);
			++errs;
		}
	}

	if (errs == 0)
	{
		FILE *fp = fopen(fname, "r");
		if (fp == NULL)
			perror(fname);
		else
		{
			struct stat st;
			char *buf = NULL;
			if (fstat(fileno(fp), &st) == 0 &&
				st.st_size < 65535 &&
				(buf = malloc(st.st_size + 1)) != NULL &&
				fread(buf, st.st_size, 1, fp) == 1)
			{
				token tok;
				memset(&tok, 0, sizeof tok);
				tok.p = buf;
				buf[st.st_size] = 0;
				if (verbose)
					printf("scanning %s\n-----\n", buf);

				rtc = a_r_parse(buf, &my_cb, NULL);
				if (verbose)
					printf("-----\na_r_parse returned %d\n", rtc);
			}

			free(buf);
			fclose(fp);
		}
	}

	return errs || rtc;
}
#elif defined TEST_LONGEST_SUBSTRING
#include <stdio.h>
int main(int argc, char *argv[])
{
	if (argc != 3) {
		printf("usage: longest-common-substring string1 string2\n");
		return 1;
	}

	printf("longest = %d\n", longest_substring(argv[1], argv[2]));

	return 0;
}
#endif // TEST_UTIL / TEST_LONGEST_SUBSTRING
